"Scope and variable tracking for Hy/Python scopes."

import ast
import itertools
from abc import ABC, abstractmethod

import hy.compat
from hy.models import Expression, List, Symbol, Tuple
from hy.reader import mangle


def is_function_scope(scope):
    return isinstance(scope, ScopeFn) and scope.is_fn


def is_inside_function_scope(scope):
    "True if any enclosing scope (including this one) is a function scope."
    cur = scope
    while not isinstance(cur, ScopeGlobal):
        if is_function_scope(cur):
            return True
        cur = cur.parent
    return False


def nearest_python_scope(scope):
    "Return the closest enclosing scope that corresponds to a Python scope."
    cur = scope
    while isinstance(cur, (ScopeLet, ScopeGen)):
        cur = cur.parent
    return cur


class OuterVar(ast.stmt):
    "Custom AST node that can compile to either Nonlocal or Global."
    def __init__(self, expr, scope, names):
        from hy.compiler import asty
        super().__init__()
        self.__dict__.update(asty._get_pos(expr))
        self._scope = scope
        self.names = names


class ResolveOuterVars(ast.NodeTransformer):
    "Find all OuterVar nodes and replace with Nonlocal or Global as necessary."
    def visit_OuterVar(self, node):
        from hy.compiler import asty
        scope = node._scope
        defined = set()
        undefined = list(node.names)  # keep order, so can't use set
        while undefined and scope.parent:
            scope = scope.parent
            has = set()
            if isinstance(scope, ScopeFn):
                has = scope.defined
            elif isinstance(scope, ScopeLet):
                has = set(scope.bindings.keys())
            elif isinstance(scope, ScopeGlobal):
                res = []
                if not scope.defined.issuperset(undefined):
                    # emit nonlocal, let python raise the error
                    break
                if undefined:
                    res.append(asty.Global(node, names=list(undefined)))
                if defined:
                    res.append(asty.Nonlocal(node, names=list(defined)))
                return res
            defined.update(has.intersection(undefined))
            undefined = [name for name in undefined if name not in has]
        return [asty.Nonlocal(node, names=node.names)] if node.names else []


class NodeRef:
    """
    Wrapper for AST nodes that have symbol names, so that we can rename them if
    necessary. Each `NodeRef` corresponds to one Python identifier. `ast` nodes
    that accept multiple identifier names (`global`, `nonlocal`, etc) have
    their specific identifier referenced by their `index` in the list of
    `names` provided by the `ast` node.
    """

    ACCESSOR = {
        ast.Name: "id",
        ast.Global: "names",
        ast.Nonlocal: "names",
        OuterVar: "names",
    }
    if hy.compat.PY3_10:
        ACCESSOR.update(
            {
                ast.MatchAs: "name",
                ast.MatchStar: "name",
                ast.MatchMapping: "rest",
            }
        )

    def __init__(self, node, index=None):
        self.node = node
        self.index = index
        self._accessor = NodeRef.ACCESSOR[type(self.node)]

    @property
    def name(self):
        res = getattr(self.node, self._accessor)
        if self.index is not None:
            return res[self.index]
        return res

    @name.setter
    def name(self, new_name):
        "Used to rename `ast` identifiers"
        if self.index is not None:
            getattr(self.node, self._accessor)[self.index] = new_name
        else:
            setattr(self.node, self._accessor, new_name)

    @staticmethod
    def wrap(f):
        "Decorator to convert AST node parameter to NodeRef."

        def _wrapper(self, node, index=None):
            if not isinstance(node, NodeRef):
                node = NodeRef(node, index)
            return f(self, node)

        return _wrapper

    def __repr__(self):
        return (
            f"NodeRef(name={self.name}, node={type(self.node).__name__},"
            f" index={self.index})"
        )


class ScopeBase(ABC):
    def __init__(self, compiler):
        self.parent = None
        self.compiler = compiler
        self.children = []

    def create(self, scope_type, *args):
        "Create new scope from this one."
        return scope_type(self.compiler, *args)

    def __enter__(self):
        if self.compiler.scope is not self:
            self.parent = self.compiler.scope
            if self not in self.parent.children:
                self.parent.children.append(self)
        self.compiler.scope = self
        return self

    def __exit__(self, *args):
        if self.parent:
            self.compiler.scope = self.parent
        return False

    # Scope interface
    @abstractmethod
    def access(self, node, index=None):
        "Called when a symbol is accessed."
        ...

    @abstractmethod
    def assign(self, node, index=None):
        "Called when a symbol is assigned to."
        ...

    @abstractmethod
    def define(self, name):
        """
        Called when a symbol is defined.
        (e.g., function names, class names, imports)
        """
        ...

    @abstractmethod
    def define_nonlocal(self, node, root):
        "Called when a symbol is declared nonlocal or global."
        ...


class ScopeGlobal(ScopeBase):
    """Global scope."""

    def __init__(self, compiler):
        super().__init__(compiler)
        self.defined = set()
        "set: of all symbols created or assigned to in this scope"

        self.nonlocal_vars = []
        """
        list: of all `nonlocal`'s defined in this scope.

        Explicitly not a `set` so that we maintain the order they were defined in
        """

    def __exit__(self, *args):
        nonlocal_vars = self.nonlocal_vars
        self.nonlocal_vars = []
        if not self.defined.issuperset(nonlocal_vars):
            raise SyntaxError(f"no binding for nonlocal '{nonlocal_vars[0]}'")
        return super().__exit__(*args)

    @NodeRef.wrap
    def access(self, node, index=0):
        return node.node

    @NodeRef.wrap
    def assign(self, node):
        self.define(node.name)
        return node.node

    def define(self, name):
        self.defined.add(name)

    def define_nonlocal(self, node, root):
        if root == "nonlocal":
            self.nonlocal_vars.extend(node.names)
        else:
            self.defined.update(node.names)


class ScopeLet(ScopeBase):
    """
    Scope that supports let-binding by renaming bound symbols as they are
    accessed/assigned. Defined symbols are never renamed.
    """

    def __init__(self, compiler):
        super().__init__(compiler)
        self.bindings = {}

    def _rename_if_bound(self, node):
        if node.name in self.bindings:
            node.name = self.bindings[node.name]
            return True
        return False

    @NodeRef.wrap
    def access(self, node):
        self._rename_if_bound(node) or self.parent.access(node)
        return node.node

    @NodeRef.wrap
    def assign(self, node):
        self._rename_if_bound(node) or self.parent.assign(node)
        return node.node

    def define(self, name):
        self.bindings.pop(name, None)
        self.parent.define(name)

    def define_nonlocal(self, node, root):
        # remove nonlocal defs of any let scopes in this Python scope
        cur = self
        while isinstance(cur, ScopeLet):
            for name in node.names:
                if root == "nonlocal":
                    if name in cur.bindings and cur is not self:
                        node.names.remove(name)
                else:
                    cur.bindings[name] = name
            cur = cur.parent
        cur.define_nonlocal(node, root)

    def add(self, target, new_name=None):
        """Add a new let-binding target, mapped to a new, unique name."""
        if isinstance(target, (str, Symbol)):
            if "." in target:
                raise ValueError("binding target may not contain a dot")
            name = mangle(target)
            if new_name is None:
                new_name = self.compiler.get_anon_var(f"_hy_let_{name}")
            self.bindings[name] = new_name
            if isinstance(target, Symbol):
                return Symbol(new_name).replace(target)
            return new_name
        if new_name is not None:
            raise ValueError("cannot specify name for compound targets")
        if isinstance(target, (List, Tuple)):
            return target.__class__(map(self.add, target)).replace(target)
        if (
            isinstance(target, Expression)
            and target
            and target[0] == Symbol("unpack-iterable")
        ):
            return Expression([target[0], *map(self.add, target[1:])]).replace(target)
        raise ValueError(f"invalid binding target: {type(target)}")


class ScopeFn(ScopeBase):
    """Scope that corresponds to Python's own function or class scopes."""

    def __init__(self, compiler, args=None, is_async=False):
        super().__init__(compiler)
        self.defined = set()
        "set: of all vars defined in this scope"
        self.seen = []
        "list: of all vars accessedto in this scope"
        self.nonlocal_vars = {}
        "set: of all `nonlocal`'s defined in this scope"
        self.is_fn = args is not None
        """
        bool: `True` if this scope is being used to track a python
        function `False` for classes
        """
        self.is_async = is_async
        """bool: `True` if this scope is for an async function,
        which may need special handling during compilation"""
        self.has_yield = False
        """bool: `True` if this scope is tracking a function that has `yield`
        statements, as generator functions may need special handling"""

        if args:
            for arg in itertools.chain(
                args.args, args.posonlyargs, args.kwonlyargs, [args.vararg, args.kwarg]
            ):
                if arg:
                    self.define(arg.arg)

    def __exit__(self, *args):
        self.defined.difference_update(self.nonlocal_vars.keys())
        for node in self.seen:
            if node.name not in self.defined:
                # pass unbound/nonlocal names up to parent scope
                self.parent.access(node)
        return super().__exit__(*args)

    @NodeRef.wrap
    def access(self, node):
        self.seen.append(node)
        return node.node

    @NodeRef.wrap
    def assign(self, node):
        self.access(node)
        self.define(node.name)
        return node.node

    def define(self, name):
        self.defined.add(name)

    def define_nonlocal(self, node, root):
        if root == "nonlocal":
            self.nonlocal_vars.update({name: node for name in node.names})
        else:
            self.defined.update(node.names)

        for n in self.seen:
            if n.name in node.names:
                raise SyntaxError(
                    f"name '{n.name}' is declared {root} after being used"
                )
        if root == "nonlocal":
            # toss all nonlocal names up to parent scope
            for i in range(len(node.names)):
                self.parent.access(node, i)


class ScopeGen(ScopeFn):
    """
    Scope that supports generator forms (`lfor`, `gfor`, ...). If this scope is
    contained within a function scope or the global scope, then any variable
    assignments are "leaked out" to the parent scope, mimicking Python's inline
    generator semantics.

    .. note::
       see :pep:`572#why-not-use-a-sublocal-scope-and-prevent-namespace-pollution`
       for more details on how and when variables are leaked into enclosing scopes.
    """

    def __init__(self, compiler):
        super().__init__(compiler)
        self.iterators = set()
        self.assignments = []
        self.exposing_assignments = False

    def __enter__(self):
        super().__enter__()
        enclosing = nearest_python_scope(self.parent)
        if isinstance(enclosing, ScopeGlobal) or is_function_scope(enclosing):
            # only "leak out" assignments if we're contained in a function
            # (or global) scope
            self.exposing_assignments = True
        return self

    def finalize(self):
        """
        Access and return all the identifiers created and potentially leaked
        out by this generator. This should be called once and only once we've
        processed all of the iterators in this generators so as not to leak out
        any unwanted identifiers
        """
        res = set()
        for node in self.assignments:
            if node.name not in self.nonlocal_vars:
                self.parent.access(node)
                res.add(node.name)
        return sorted(res)

    @NodeRef.wrap
    def assign(self, node):
        self.access(node)
        if node.name not in self.defined:
            self.assignments.append(node)
        return node.node

    @NodeRef.wrap
    def access(self, node):
        if not node.name in self.iterators:
            self.seen.append(node)
        return node.node

    def iterator(self, target):
        """
        Declare an iteration variable name for this scope; as in Python, the
        iteration variable(s) cannot be reassigned.
        """
        self.iterators.update(
            name.id for name in ast.walk(target) if isinstance(name, ast.Name)
        )
        # remove potentially leakable identifiers that were actually iteration
        # variables found in `target`
        self.assignments = [
            node for node in self.assignments if node.name not in self.iterators
        ]
        self.seen = [node for node in self.seen if node.name not in self.iterators]



import builtins
import importlib
import inspect
import os
import re
import sys
import traceback
from ast import AST

from funcparserlib.parser import NoParseError

import hy.compiler
from hy.compat import PY3_11
from hy.errors import (
    HyLanguageError,
    HyMacroExpansionError,
    HyRequireError,
    HyTypeError,
)
from hy.model_patterns import whole
from hy.models import Expression, Symbol, as_model, is_unpack, replace_hy_obj
from hy.reader import mangle
from hy.reader.mangling import slashes2dots

EXTRA_MACROS = ["hy.core.result_macros", "hy.core.macros"]


def macro(name):
    """Decorator to define a macro called `name`."""
    return lambda fn: install_macro(name, fn, fn)


def reader_macro(name, fn):
    fn = rename_function(fn, name)
    fn.__globals__.setdefault("_hy_reader_macros", {})[name] = fn


def pattern_macro(names, pattern, shadow=None):
    pattern = whole(pattern)
    py_version_required = None
    if isinstance(names, tuple):
        py_version_required, names = names

    def dec(fn):
        def wrapper_maker(name):
            def wrapper(_hy_compiler, *args):

                if shadow and any(is_unpack("iterable", x) for x in args):
                    # Try a shadow function call with this name instead.
                    return Expression(
                        [Expression(map(Symbol, [".", "hy", "pyops", name])), *args]
                    ).replace(_hy_compiler.this)

                expr = _hy_compiler.this

                if py_version_required and sys.version_info < py_version_required:
                    raise _hy_compiler._syntax_error(
                        expr,
                        "`{}` requires Python {} or later".format(
                            name, ".".join(map(str, py_version_required))
                        ),
                    )

                try:
                    parse_tree = pattern.parse(args)
                except NoParseError as e:
                    raise _hy_compiler._syntax_error(
                        expr[min(e.state.pos + 1, len(expr) - 1)],
                        "parse error for pattern macro '{}': {}".format(
                            name, e.msg.replace("end of input", "end of macro call")
                        ),
                    )
                return fn(_hy_compiler, expr, name, *parse_tree)

            return wrapper

        for name in [names] if isinstance(names, str) else names:
            install_macro(name, wrapper_maker(name), fn)
        return fn

    return dec


def install_macro(name, fn, module_of):
    name = mangle(name)
    fn = rename_function(fn, name)
    module_of.__globals__.setdefault("_hy_macros", {})[name] = fn
    return fn


def _same_modules(source_module, target_module):
    """Compare the filenames associated with the given modules names.

    This tries to not actually load the modules.
    """
    if not (source_module or target_module):
        return False

    if target_module is source_module:
        return True

    def get_filename(module):
        if inspect.ismodule(module):
            return inspect.getfile(module)
        elif (
                (spec := importlib.util.find_spec(module)) and
                isinstance(spec.loader, importlib.machinery.SourceFileLoader)):
            return spec.loader.get_filename()

    try:
        return os.path.samefile(
            get_filename(source_module),
            get_filename(target_module))
    except (ValueError, TypeError, ImportError, FileNotFoundError):
        return False


def derive_target_module(target_module, parent_frame):
    if target_module is None:
        target_namespace = parent_frame.f_globals
        target_module = target_namespace.get("__name__", None)
    elif isinstance(target_module, str):
        target_module = importlib.import_module(target_module)
        target_namespace = target_module.__dict__
    elif inspect.ismodule(target_module):
        target_namespace = target_module.__dict__
    else:
        raise HyTypeError(
            "`target_module` is not a recognized type: {}".format(type(target_module))
        )
    return target_module, target_namespace


def import_module_from_string(module_name, package_module):
    package = None
    if module_name.startswith("."):
        source_dirs = module_name.split(".")
        target_dirs = getattr(package_module, "__name__", package_module).split(".")
        while len(source_dirs) > 1 and source_dirs[0] == "" and target_dirs:
            source_dirs.pop(0)
            target_dirs.pop()
        package = ".".join(target_dirs + source_dirs[:-1])
    try:
        return importlib.import_module(module_name, package)
    except ImportError as e:
        raise HyRequireError(e.args[0]).with_traceback(None)


def require_reader(source_module, target_module, assignments):
    target_module, target_namespace = derive_target_module(
        target_module, inspect.stack()[1][0]
    )

    if _same_modules(source_module, target_module):
        return False

    if not inspect.ismodule(source_module):
        source_module = import_module_from_string(source_module, target_module)

    source_macros = source_module.__dict__.setdefault("_hy_reader_macros", {})
    target_macros = target_namespace.setdefault("_hy_reader_macros", {})

    assignments = (
        source_macros.keys() if assignments == "ALL" else assignments
    )

    for name in assignments:
        if name in source_module._hy_reader_macros:
            target_macros[name] = source_macros[name]
        else:
            raise HyRequireError(f"Could not require name {name} from {source_module}")

    return True


def enable_readers(module, reader, names):
    _, namespace = derive_target_module(module, inspect.stack()[1][0])
    names = (
        namespace["_hy_reader_macros"].keys() if names == "ALL" else names
    )
    for name in names:
        if name not in namespace["_hy_reader_macros"]:
            raise NameError(f"reader {name} is not defined")
        reader.reader_macros[name] = namespace["_hy_reader_macros"][name]


def require(source_module, target, assignments, prefix="", target_module_name=None, compiler=None):
    """Load macros from a module. Return a list of (new name, source
    name, macro object) tuples, including only macros that were
    actually transferred.

    - `target` can be a a string (naming a module), a module object,
      a dictionary, or `None` (meaning the calling module).
    - `assignments` can be "ALL", "EXPORTS", or a list of (macro
      name, alias) pairs."""

    if type(target) is dict:
        target_module = None
    else:
        target_module, target_namespace = derive_target_module(
            target, inspect.stack()[1][0]
        )
        # Let's do a quick check to make sure the source module isn't actually
        # the module being compiled (e.g. when `runpy` executes a module's code
        # in `__main__`).
        # We use the module's underlying filename for this (when they exist), since
        # it's the most "fixed" attribute.
        if _same_modules(source_module, target_module):
            return []

    if not inspect.ismodule(source_module):
        source_module = import_module_from_string(source_module,
           target_module_name or target_module or '')

    source_macros = source_module.__dict__.setdefault("_hy_macros", {})
    source_exports = getattr(
        source_module,
        "_hy_export_macros",
        [k for k in source_macros.keys() if not k.startswith("_")],
    )

    if not source_module._hy_macros:
        if assignments in ("ALL", "EXPORTS"):
            return []
        out = []
        for name, alias in assignments:
            try:
                out.extend(require(
                    f"{source_module.__name__}.{mangle(name)}",
                    target_module or target,
                    "ALL",
                    prefix=alias,
                ))
            except HyRequireError as e:
                raise HyRequireError(
                    f"Cannot import name '{name}'"
                    f" from '{source_module.__name__}'"
                    f" ({source_module.__file__})"
                )
        return out

    target_macros = target_namespace.setdefault("_hy_macros", {}) if target_module else target

    if prefix:
        prefix += "."
    out = []

    for name, alias in (
        assignments
        if assignments not in ("ALL", "EXPORTS")
        else (
            (k, k)
            for k in source_macros.keys()
            if assignments == "ALL" or k in source_exports
        )
    ):
        _name = mangle(name)
        if compiler:
            compiler.warn_on_core_shadow(prefix + alias)
        alias = mangle(prefix + alias)
        if _name in source_module._hy_macros:
            target_macros[alias] = source_macros[_name]
            out.append((alias, _name, source_macros[_name]))
        else:
            raise HyRequireError(
                "Could not require name {} from {}".format(_name, source_module)
            )

    return out


def require_vals(*args, **kwargs):
    return [value for _, _, value in require(*args, **kwargs)]


def local_macro_name(original):
    return '_hy_local_macro__' + (mangle(original)
      # We have to do a bit of extra mangling beyond `mangle` itself to
      # handle names with periods.
        .replace('D', 'DN')
        .replace('.', 'DD'))


def load_macros(module):
    """Load the hy builtin macros into module `module_name`,
    removing any prior macros set.
    It is an error to call this on any module in `hy.core`.
    """
    builtin_macros = EXTRA_MACROS
    module._hy_macros = {}
    module._hy_reader_macros = {}

    for builtin_mod_name in builtin_macros:
        builtin_mod = importlib.import_module(builtin_mod_name)

        # This may overwrite macros in the module.
        if hasattr(builtin_mod, "_hy_macros"):
            module._hy_macros.update(getattr(builtin_mod, "_hy_macros", {}))

        if hasattr(builtin_mod, "_hy_reader_macros"):
            module._hy_reader_macros.update(
                getattr(builtin_mod, "_hy_reader_macros", {})
            )


class MacroExceptions:
    """wrap non ``HyLanguageError``'s in ``HyMacroExpansionError`` preserving stack trace

    used in lieu of ``@contextmanager`` to ensure stack trace contains only internal hy
    modules for consistent filtering.
    """

    def __init__(self, module, macro_tree, compiler=None):
        self.module = module
        self.macro_tree = macro_tree
        self.compiler = compiler

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is None:
            return True
        elif not issubclass(exc_type, HyLanguageError):
            if self.compiler:
                filename = self.compiler.filename
                source = self.compiler.source
            else:
                filename = None
                source = None

            exc_msg = "  ".join(
                traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])
            )

            msg = "expanding macro {}\n  ".format(str(self.macro_tree[0]))
            msg += exc_msg

            raise HyMacroExpansionError(msg, self.macro_tree, filename, source)
        else:
            return False


def macroexpand(tree, module, compiler=None, once=False, result_ok=True):
    '''If `tree` isn't an `Expression` that might be a macro call,
    return it unchanged. Otherwise, try to expand it. Do this
    repeatedly unless `once` is true. Call `as_model` after each
    expansion. If the return value is a compiler `Result` object, and
    `result_ok` is false, return the previous value. Otherwise, return
    the final expansion.'''

    if not inspect.ismodule(module):
        module = importlib.import_module(module)

    assert not compiler or compiler.module == module

    while isinstance(tree, Expression) and tree:

        fn = tree[0]
        if (
                isinstance(fn, Expression) and
                fn and
                fn[0] == Symbol(".") and
                all(isinstance(x, Symbol) for x in fn)):
            fn = ".".join(map(mangle, fn[1:]))
        elif isinstance(fn, Symbol):
            fn = mangle(fn)
        else:
            break

        if fn.startswith('hy.R.'):
            # Special syntax for a one-shot `require`.
            req_from, _, fn = fn[len('hy.R.'):].partition('.')
            req_from = slashes2dots(req_from)
            try:
                m = importlib.import_module(req_from)._hy_macros[fn]
            except ImportError as e:
                raise HyRequireError(e.args[0]).with_traceback(None)
            except (AttributeError, KeyError):
                raise HyRequireError(f'Could not require name {fn} from {req_from}')
        else:
            # Choose the first namespace with the macro.
            m = ((compiler and next(
                    (d[fn]
                        for d in [
                            compiler.extra_macros,
                            *(s['macros'] for s in reversed(compiler.local_state_stack))]
                        if fn in d),
                    None)) or
                next(
                    (mod._hy_macros[fn]
                        for mod in (module, builtins)
                        if fn in getattr(mod, "_hy_macros", ())),
                    None))
            if not m:
                break

        with MacroExceptions(module, tree, compiler):
            if compiler:
                compiler.this = tree
            obj = m(
                # If the macro's first parameter is named
                # `_hy_compiler`, pass in the current compiler object
                # in its place.
                *([compiler]
                    if m.__code__.co_varnames[:1] == ('_hy_compiler',)
                    else []),
                *map(as_model, tree[1:]))
            if isinstance(obj, (hy.compiler.Result, AST)):
                return obj if result_ok else tree

            tree = replace_hy_obj(obj, tree)

        if once:
            break

    return tree


def rename_function(f, new_name):
    """Create a copy of a function, but with a new name."""
    f = type(f)(
        f.__code__.replace(
            co_name=new_name,
            **(
                {
                    "co_qualname": re.sub(
                        r"\.[^.+]\Z", "." + new_name, f.__code__.co_qualname
                    )
                    if "." in f.__code__.co_qualname
                    else new_name
                }
                if PY3_11
                else {}
            ),
        ),
        f.__globals__,
        str(new_name),
        f.__defaults__,
        f.__closure__,
    )
    f.__dict__.update(f.__dict__)
    return f



import re
import unicodedata

MANGLE_DELIM = "X"

normalizes_to_underscore = "_︳︴﹍﹎﹏＿"


def mangle(s):
    """Stringify the argument (with :class:`str`, not :func:`repr` or
    :hy:func:`hy.repr`) and convert it to a valid Python identifier according
    to :ref:`Hy's mangling rules <mangling>`. ::

        (hy.mangle 'foo-bar)   ; => "foo_bar"
        (hy.mangle "🦑")       ; => "hyx_XsquidX"

    If the stringified argument is already both legal as a Python identifier
    and normalized according to Unicode normalization form KC (NFKC), it will
    be returned unchanged. Thus, ``hy.mangle`` is idempotent. ::

        (setv x '♦-->♠)
        (= (hy.mangle (hy.mangle x)) (hy.mangle x))  ; => True

    Generally, the stringifed input is expected to be parsable as a symbol. As
    a convenience, it can also have the syntax of a :ref:`dotted identifier
    <dotted-identifiers>`, and ``hy.mangle`` will mangle the dot-delimited
    parts separately. ::

        (hy.mangle "a.c!.d")  ; => "a.hyx_cXexclamation_markX.d"
    """

    assert s
    s = str(s)

    if "." in s and s.strip("."):
        return ".".join(mangle(x) if x else "" for x in s.split("."))

    # Remove and save leading underscores
    s2 = s.lstrip(normalizes_to_underscore)
    leading_underscores = "_" * (len(s) - len(s2))
    s = s2

    # Convert hyphens without introducing a new leading underscore
    s = s[0] + s[1:].replace("-", "_") if s else s

    # Convert invalid characters or reserved words
    if not (leading_underscores + s).isidentifier():
        # Replace illegal characters with their Unicode character
        # names, or hexadecimal if they don't have one.
        s = "hyx_" + "".join(
            c if c != MANGLE_DELIM and ("S" + c).isidentifier()
            # We prepend the "S" because some characters aren't
            # allowed at the start of an identifier.
            else "{0}{1}{0}".format(
                MANGLE_DELIM,
                unicodedata.name(c, "").lower().replace("-", "H").replace(" ", "_")
                or "U{:x}".format(ord(c)),
            )
            for c in s
        )

    # Step 5: Add back leading underscores
    s = leading_underscores + s

    # Normalize Unicode per PEP 3131.
    s = unicodedata.normalize("NFKC", s)

    assert s.isidentifier()
    return s


def unmangle(s):
    """Stringify the argument and try to convert it to a pretty unmangled
    form. See :ref:`Hy's mangling rules <mangling>`. ::

        (hy.unmangle "hyx_XsquidX")  ; => "🦑"

    Unmangling may not round-trip, because different Hy symbol names can mangle
    to the same Python identifier. In particular, Python itself already
    considers distinct strings that have the same normalized form (according to
    NFKC), such as ``hello`` and ``𝔥𝔢𝔩𝔩𝔬``, to be the same identifier.

    It's an error to call ``hy.unmangle`` on something that looks like a
    properly mangled name but isn't. For example, ``(hy.unmangle
    "hyx_XpizzazzX")`` is erroneous, because there is no Unicode character
    named "PIZZAZZ" (yet)."""

    s = str(s)

    prefix = ""
    suffix = ""
    m = re.fullmatch(r"(_+)(.*?)(_*)", s, re.DOTALL)
    if m:
        prefix, s, suffix = m.groups()

    if s.startswith("hyx_"):
        s = re.sub(
            "{0}(U)?([_a-z0-9H]+?){0}".format(MANGLE_DELIM),
            lambda mo: chr(int(mo.group(2), base=16))
            if mo.group(1)
            else unicodedata.lookup(
                mo.group(2).replace("_", " ").replace("H", "-").upper()
            ),
            s[len("hyx_") :],
        )
    s = s.replace("_", "-")

    return prefix + s + suffix


def slashes2dots(s):
    'Interpret forward slashes as a substitute for periods.'
    return mangle(re.sub(
        r'/(-*)',
        lambda m: '.' + '_' * len(m.group(1)),
        unmangle(s)))



__version__ = '1.2.0'
nickname = 'Crackers and Snacks'
last_version = '1.2.0'
  # This is used by `(pragma :hy …)` to guess whether an unreleased
  # version of Hy is new enough. In a released version, it's simply
  # equal to `__version__`.


def _initialize_env_var(env_var, default_val):
    import os

    return bool(os.environ.get(env_var, default_val))


# Import for side-effects.
import hy.importer, hy.hy_inspect
hy.importer._inject_builtins()



class I:
    """``hy.I`` is an object that provides syntactic sugar for imports. It allows syntax like ``(hy.I.math.sqrt 2)``  to mean ``(import math) (math.sqrt 2)``, except without bringing ``math`` or ``math.sqrt`` into scope. (See :ref:`hy.R <hy.R>` for a version that requires a macro instead of importing a Python object.) This is useful in macros to avoid namespace pollution. To refer to a module with dots in its name, use slashes instead: ``hy.I.os/path.basename`` gets the function ``basename`` from the module ``os.path``.

    You can also call ``hy.I`` like a function, as in ``(hy.I "math")``, which is useful when the module name isn't known until run-time. This interface just calls :py:func:`importlib.import_module`, avoiding (1) mangling due to attribute lookup, and (2) the translation of ``/`` to ``.`` in the module name. The advantage of ``(hy.I modname)`` over ``importlib.import_module(modname)`` is merely that it avoids bringing ``importlib`` itself into scope."""
    def __call__(self, module_name):
        import importlib
        return importlib.import_module(module_name)
    def __getattr__(self, s):
        from hy.reader.mangling import slashes2dots
        return self(slashes2dots(s))
I = I()


# Import some names on demand so that the dependent modules don't have
# to be loaded if they're not needed.

_jit_imports = dict(
    pyops=["hy.pyops", None],
    read="hy.reader",
    read_many="hy.reader",
    mangle="hy.reader",
    unmangle="hy.reader",
    eval=["hy.compiler", "hy_eval_user"],
    repr=["hy.core.hy_repr", "hy_repr"],
    repr_register=["hy.core.hy_repr", "hy_repr_register"],
    gensym="hy.core.util",
    macroexpand="hy.core.util",
    macroexpand_1="hy.core.util",
    disassemble="hy.core.util",
    as_model="hy.models",
    REPL="hy.repl",
    Reader="hy.reader.reader",
    HyReader="hy.reader.hy_reader",
    PrematureEndOfInput="hy.reader.exceptions"
)


def __getattr__(k):
    if k not in _jit_imports:
        raise AttributeError(f"module {__name__!r} has no attribute {k!r}")

    v = _jit_imports[k]
    module, original_name = v if isinstance(v, list) else (v, k)
    import importlib
    module = importlib.import_module(module)

    globals()[k] = getattr(module, original_name) if original_name else module
    return globals()[k]
